// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <atomic>
#include <errno.h>
#include <jansson.h>
#include <poll.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <string>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <vector>

#include "source.h"
#include "filter.h"
#include "cfg.h"
#include "utils.h"
#include "log.h"
#include "error.h"
#include "motion_trigger.h"
#include "http_server_support.h"

std::string run_rest(configuration_t *const cfg, const std::string & path, const std::string & pars, const std::string & snapshot_dir, const int quality)
{
	int code = 200;
	std::vector<std::string> *parts = split(path, "/");

	json_t *json = json_object();

	bool cmd = parts -> at(0) == "cmd";

	cfg -> lock.lock();
	interface *i = (parts -> size() >= 1 && !cmd) ? find_interface_by_id(cfg, parts -> at(0)) : nullptr;

	if (parts -> size() < 2) {
		code = 500;
		json_object_set_new(json, "msg", json_string("ID/cmd missing"));
		json_object_set_new(json, "result", json_false());
	}
	else if (cmd && parts -> at(1) == "list") {
		json_object_set_new(json, "msg", json_string("OK"));
		json_object_set_new(json, "result", json_true());

		json_t *out_arr = json_array();
		json_object_set_new(json, "data", out_arr);

		for(instance_t * inst : cfg -> instances) {
			for(interface *cur : inst -> interfaces) {
				json_t *record = json_object();

				json_object_set_new(record, "id", json_string(cur -> get_id().c_str()));

				json_object_set_new(record, "id-descr", json_string(cur -> get_descr().c_str()));

				switch(cur -> get_class_type()) {
					case CT_HTTPSERVER:
						json_object_set_new(record, "type", json_string("http-server"));
						break;
					case CT_MOTIONTRIGGER:
						json_object_set_new(record, "type", json_string("motion-trigger"));
						break;
					case CT_TARGET:
						json_object_set_new(record, "type", json_string("stream-writer"));
						break;
					case CT_SOURCE:
						json_object_set_new(record, "type", json_string("source"));
						break;
					case CT_LOOPBACK:
						json_object_set_new(record, "type", json_string("video4linux-loopback"));
						break;
					default:
						json_object_set_new(record, "type", json_string("INTERNAL ERROR"));
						break;
				}

				json_object_set_new(record, "running", json_string(cur -> is_paused() ? "false" : "true"));

				json_object_set_new(record, "enabled", json_string(cur -> is_running() ? "true" : "false"));

				json_array_append_new(out_arr, record);
			}
		}
	}
	// commands that require an ID
	else if (i == nullptr) {
		code = 404;
		json_object_set_new(json, "msg", json_string("ID not found"));
		json_object_set_new(json, "result", json_false());
	}
	else if (cmd && parts -> at(1) == "list-files" && i -> get_class_type() == CT_TARGET) {
		json_object_set_new(json, "msg", json_string("OK"));
		json_object_set_new(json, "result", json_true());

		json_t *out_arr = json_array();
		json_object_set_new(json, "data", out_arr);

		auto *files = load_filelist(((target *)i) -> get_target_dir(), "");

		for(auto file : *files) {
			json_t *record = json_object();

			json_object_set_new(record, "file", json_string(file.name.c_str()));
			json_object_set_new(record, "mtime", json_integer(file.last_change));
			json_object_set_new(record, "size", json_integer(file.size));

			json_array_append_new(out_arr, record);
		}

		delete files;
	}
	else if (cmd && parts -> at(1) == "take-a-snapshot" && i -> get_class_type() == CT_SOURCE) {
		if (take_a_picture(static_cast<source *>(i), snapshot_dir, quality, false)) {
			json_object_set_new(json, "msg", json_string("OK"));
			json_object_set_new(json, "result", json_true());
		}
		else {
			json_object_set_new(json, "msg", json_string("failed"));
			json_object_set_new(json, "result", json_false());
		}
	}
	else if (cmd && parts -> at(1) == "start-a-recording" && i -> get_class_type() == CT_SOURCE) {
		const bool is_view_proxy = strstr(pars.c_str(), "view-proxy") != nullptr;

		interface *str = start_a_video(static_cast<source *>(i), snapshot_dir, quality, cfg, is_view_proxy);

		if (str) {
			json_object_set_new(json, "msg", json_string("OK"));
			json_object_set_new(json, "result", json_true());

			find_instance_by_interface(cfg, i) -> interfaces.push_back(str);
		}
		else {
			json_object_set_new(json, "msg", json_string("failed"));
			json_object_set_new(json, "result", json_false());
		}

	}
	else if (parts -> at(1) == "pause") {
		if (i -> pause()) {
			json_object_set_new(json, "msg", json_string("OK"));
			json_object_set_new(json, "result", json_true());
		}
		else {
			json_object_set_new(json, "msg", json_string("failed"));
			json_object_set_new(json, "result", json_false());
		}
	}
	else if (parts -> at(1) == "unpause") {
		i -> unpause();
		json_object_set_new(json, "msg", json_string("OK"));
		json_object_set_new(json, "result", json_true());
	}
	else if (parts -> at(1) == "stop" || parts -> at(1) == "start") {
		if (start_stop(i, parts -> at(1) == "start")) {
			json_object_set_new(json, "msg", json_string("OK"));
			json_object_set_new(json, "result", json_true());
		}
		else {
			json_object_set_new(json, "msg", json_string("failed"));
			json_object_set_new(json, "result", json_false());
		}
	}
	else if (parts -> at(1) == "configure-motion-trigger" && parts -> size() == 4) {
		json_decref(json);
		json = ((motion_trigger *)i) -> rest_set_parameter(parts -> at(2), parts -> at(3));
	}
	else if (parts -> at(1) == "list-motion-trigger") {
		json_t *result = ((motion_trigger *)i) -> get_rest_parameters();

		json_object_set_new(json, "msg", json_string("OK"));
		json_object_set_new(json, "result", json_true());
		json_object_set_new(json, "data", result);
	}
	else {
		code = 404;
	}

	cfg -> lock.unlock();

	delete parts;

	char *js = json_dumps(json, JSON_COMPACT);
	std::string reply = myformat("HTTP/1.0 %d\r\nServer: " NAME " " VERSION "\r\nContent-Type: application/json\r\n\r\n%s", code, js);
	free(js);

	json_decref(json);

	return reply;
}

void handle_rest(int cfd, configuration_t *const cfg, std::atomic_bool *const global_stopflag, const std::string & snapshot_dir, const int quality)
{
	sigset_t all_sigs;
	sigfillset(&all_sigs);
	pthread_sigmask(SIG_BLOCK, &all_sigs, nullptr);

	char request_headers[65536]{ 0 };
	int h_n = 0;

	struct pollfd fds[1] { { cfd, POLLIN, 0 } };

	for(;!*global_stopflag;) {
		if (poll(fds, 1, 250) == 0)
			continue;

		int rc = read(cfd, &request_headers[h_n], sizeof request_headers - h_n - 1);
		if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN) {
				log(LL_DEBUG, "EINTR/EAGAIN %d", errno);
				continue;
			}

			log(LL_DEBUG, "error receiving request headers");
			close(cfd);
			return;
		}
		else if (rc == 0)
		{
			log(LL_DEBUG, "error receiving request headers");
			close(cfd);
			return;
		}

		h_n += rc;

		request_headers[h_n] = 0x00;

		if (strstr(request_headers, "\r\n\r\n") || strstr(request_headers, "\n\n"))
			break;
	}

	char *Path = nullptr;

	if (strncmp(request_headers, "GET ", 4) == 0)
		Path = strdup(&request_headers[4]);
	else if (strncmp(request_headers, "HEAD ", 5) == 0)
		Path = strdup(&request_headers[5]);
	else {
		close(cfd);
		return;
	}

	char *path = Path;
	while(*path == '/')
		path++;

	char *pars = nullptr;

	{
		char *dummy = strchr(path, '\r');
		if (!dummy)
			dummy = strchr(path, '\n');
		if (dummy)
			*dummy = 0x00;

		dummy = strchr(path, ' ');
		if (dummy)
			*dummy = 0x00;

		log(LL_DEBUG, "URL: %s", path);

		char *q = strchr(path, '?');
		if (q) {
			*q = 0x00;

			pars = un_url_escape(q + 1);
		}
	}

	if (strncmp(path, "rest/", 5) == 0) {
		log(LL_DEBUG, "%s", path);
		std::string reply = run_rest(cfg, &path[5], pars ? pars : "", snapshot_dir, quality);

		if (WRITE(cfd, reply.c_str(), reply.size()) <= 0)
			log(LL_DEBUG, "short write on response");
	}
	else {
		log(LL_INFO, "Path %s not found", path);

		if (WRITE(cfd, "HTTP/1.0 404\r\n\r\nReliquis non loquor", 4) <= 0)
			log(LL_DEBUG, "short write on response header");
	}

	free(Path);
	free(pars);

	close(cfd);
}
